{******************************************************************************}
{                                                                              }
{                   Delphi - Component for CodeBase 5.1                        }
{       																							  }
{               ***  Tested only with CodeBasic 5.1  ***                       }
{                                                                              }
{    Not yet supported:  Memofields, Create File, Create Index, Relations      }
{                                                                              }
{  Send any comments to Hans Guenther Mischke      Fax: +49 2391 9544 44       }
{                       Brockhauser Weg 12a                                    }
{                       D-58840 Plettenberg        CIS: 100024,1541            }
{                       Germany                                                }
{******************************************************************************}

unit CBServer;

interface

uses
	SysUtils, Classes, DsgnIntf;

const
	DLLName = 'v4mdx';					{ change this for v4cdx or v4ntx }
	cbNoServer = -9;

var
	NullStr: array[0..255] of Char;	{ Array of Char for C-Functions }

type
	TDateFormat = (dtGerman, dtAmerican, dtBritish, dtNone);
	TCBServer = class(TComponent)
	private
		CODE4:longint;						{ Pointer to CODE4-Struktur }
		FAutoOpen:boolean;            {*}
		FCreateError:boolean;         {*}
		FDateFormat:TDateFormat;      {*}
		FFieldNameError:boolean;      {*}
		FGoError:boolean;             {*}
		FOpenError:boolean;           {*}
		FSafety:boolean;              {*}
		FSkipError:boolean;           {*}
		procedure Loaded; override;
		function  GetCODE4:longint;

	public
		constructor Create(AOwner:TComponent); override;

	published
		property AutoOpen:boolean read FAutoOpen write FAutoOpen default True;
		property CreateError:boolean read FCreateError write FCreateError default True;
		property DateFormat:TDateFormat read FDateFormat write FDateFormat;
		property FieldNameError:boolean read FFieldNameError write FFieldNameError default True;
		property GoError:boolean read FGoError write FGoError default True;
		property OpenError:boolean read FOpenError write FOpenError default True;
		property Safety:boolean read FSafety write FSafety default True;
		property SkipError:boolean read FSkipError write FSkipError default True;

	end;

type
	TCBFile = class(TComponent)
	private
		FServer:TCBServer;				{ Name of Codebase-Servers }
		FilePtr:longint;					{ Pointer to CODE4-Struktur of CodeBase }
		FReadLock:boolean;
		FReadOnly:boolean;
		FExclusive:boolean;
		FFileName:string;
		FAlias:string;
		FIndex:string;
		function  GetFieldPtr(Field:string):longint;
		procedure SetFileName(Value:string);
		procedure SetAllProperties;
		function  fTagPtr(TagName:string):longint;

	public
		BOF:boolean;						{ Property: Bottom of File (BOF) }
		Deleted:boolean;					{ Property: Record is Deleted }
		EOF:boolean;						{ Property: End of File (EOF) }
		Fields:integer;					{ Property: Number of Fields in File }
		Locked:boolean;					{ Property: Record is Locked }
		RecCount:longint;					{ Property: Number of Records }
		RecNum:longint;					{ Property: Recordnumber }

		constructor Create(AOwner:TComponent); override;
		{ File Methoden }
		function  Append:integer;
		function  Close:integer;
		procedure Delete;
		function  GetAlias:string;
		function  GetDATA4:longint;
		function  GetError:integer;
		function  GetPosition:integer;
		function  Go(RecNum:longint):integer;
		function  GoBottom:integer;
		function  GoEof:integer;
		function  GoTop:integer;
		function  FileLock:integer;
		function  Open:integer;
		function  Pack:integer;
		procedure Recall;
		function  RecLock:integer;
		function  ReIndex:integer;
		function  Seek(Value:string):boolean;
		function  SetPosition(Position:double):integer;
		function  Skip(NumRecs:longint):integer;
		function  Zap(StartRec:longint; EndRec:longint):integer;
		{ Index(Tag) Methoden }
		function  GetTag:string;
		function  GetTagPtr:longint;
		function  GetNextTagPtr(onTag:longint):longint;
		function  GetPrevTagPtr(onTag:longint):longint;
		function  SetTag(TagName:string):boolean;
		function  TagDescending(TagName:string):boolean;
		function  TagExpr(TagName:string):string;
		function  TagFilter(TagName:string):string;
		function  TagKeyLen(TagName:string):integer;
		function  TagUnique(TagName:string):boolean;
		function  GetTagUniqueError(TagName:string):boolean;
		procedure SetTagUniqueError(TagName:string;Value:boolean);
		{ Field Methoden }
		function  FieldDec(Field:string):byte;
		function  FieldLen(Field:string):byte;
		function  FieldType(Field:string):byte;
		function  GetBoolean(Field:string):boolean;
		function  GetDouble(Field:string):double;
		function  GetInt(Field:string):integer;
		function  GetLong(Field:string):longint;
		function  GetStr(Field:string):string;
		procedure SetBoolean(Field:string;Value:boolean);
		procedure SetDouble(Field:string;Value:double);
		procedure SetInt(Field:string;Value:integer);
		procedure SetLong(Field:string;Value:longint);
		procedure SetStr(Field:string;Value:string);

	published
		property Server:TCBServer read FServer write FServer;
		property Exclusive:boolean read FExclusive write FExclusive default False;
		property ReadLock:boolean read FReadLock write FReadLock default False;
		property ReadOnly:boolean read FReadOnly write FReadOnly default False;
		property FileName:string read FFileName write SetFileName;
		property Alias:string read FAlias write FAlias;
		property IndexTag:string read FIndex write FIndex;

	end;


implementation

	{ Prototypes of CodeBase Functions }
	function  d4init_v:longint; far; external DLLName;
	function  d4initUndo(CODE4:longint):integer; far; external DLLName;
	function  code4autoOpen(CODE4:longint; Value:integer):integer; far; external DLLName;
	function  code4createError(CODE4:longint; Value:integer):integer; far; external DLLName;
	function  code4dateFormat(CODE4:longint; Value:string):longint; far; external DLLName;
	function  code4exclusive(CODE4:longint; Value:integer):integer; far; external DLLName;
	function  code4errorCode(CODE4:longint; Value:integer):integer; far; external DLLName;
	function  code4fieldNameError(CODE4:longint; Value:integer):integer; far; external DLLName;
	function  code4goError(CODE4:longint; Value:integer):integer; far; external DLLName;
	function  code4openError(CODE4:longint; Value:integer):integer; far; external DLLName;
	function  code4readLock(CODE4:longint; Value:integer):integer; far; external DLLName;
	function  code4readOnly(CODE4:longint; Value:integer):integer; far; external DLLName;
	function  code4safety(CODE4:longint; Value:integer):integer; far; external DLLName;
	function  code4skipError(CODE4:longint; Value:integer):integer; far; external DLLName;

	function  d4alias(DATA4:longint):PChar; far; external DLLName;
	procedure d4alias_Set(DATA4:longint; Alias:PChar); far; external DLLName;
	function  d4append_blank(DATA4:longint):integer; far; external DLLName;
	function  d4bof(DATA4:longint):integer; far; external DLLName;
	function  d4bottom(DATA4:longint):integer; far; external DLLName;
	function  d4close(DATA4:longint):integer; far; external DLLName;
	procedure d4delete(DATA4:longint); far; external DLLName;
	function  d4deleted(DATA4:longint):integer; far; external DLLName;
	function  d4eof(DATA4:longint):integer; far; external DLLName;
	function  d4field(DATA4:longint; FieldName:PChar):longint; far; external DLLName;
	function  d4field_J(DATA4:longint; FieldOrd:integer):longint; far; external DLLName;
	function  d4go(DATA4:longint; RecNum:longint):integer; far; external DLLName;
	function  d4goEof(DATA4:longint):integer; far; external DLLName;
	function  d4lock(DATA4:longint; RecNum:longint):integer; far; external DLLName;
	function  d4lock_file(DATA4:longint):integer; far; external DLLName;
	function  d4lock_test(DATA4:longint; RecNum:longint):integer; far; external DLLName;
	function  d4memo_compress(DATA4:longint):integer; far; external DLLName;
	function  d4num_fields(DATA4:longint):integer; far; external DLLName;

	function  d4open(CODE4:longint; FileName:PChar):longint; far; external DLLName;
	function  d4pack(CODE4:longint):integer; far; external DLLName;
	function  d4position(DATA4:longint):double; far; external DLLName;
	function  d4position_set(DATA4:longint; Position:double):integer; far; external DLLName;
	procedure d4recall(DATA4:longint); far; external DLLName;
	function  d4reccount(DATA4:longint):longint; far; external DLLName;
	function  d4recno(DATA4:longint):longint; far; external DLLName;
	function  d4reindex(DATA4:longint):integer; far; external DLLName;
	function  d4seek(DATA4:longint; Value:PChar):integer; far; external DLLName;
	function  d4skip(DATA4:longint; NumRecs:longint):integer; far; external DLLName;
	function  d4tag(DATA4:longint; TagName:PChar):longint; far; external DLLName;
	function  d4tag_Default(DATA4:longint):longint; far; external DLLName;
	function  d4tag_Next(DATA4:longint;tagOn:longint):longint; far; external DLLName;
	function  d4tag_Prev(DATA4:longint;tagOn:longint):longint; far; external DLLName;
	procedure d4tag_Select(DATA4:longint; TagPtr:longint); far; external DLLName;
	function  d4tag_Selected(DATA4:longint):longint; far; external DLLName;
	function  d4top(DATA4:longint):integer; far; external DLLName;
	function  d4zap(DATA4:longint;StartRec:longint;EndRec:longint):integer; far; external DLLName;

	procedure f4assign(FieldPtr:longint;Value:PChar); far; external DLLName;
	procedure f4assign_Char(FieldPtr:longint;Value:integer); far; external DLLName;
	procedure f4assign_Double(FieldPtr:longint;Value:double); far; external DLLName;
	procedure f4assign_Int(FieldPtr:longint;Value:integer); far; external DLLName;
	procedure f4assign_Long(FieldPtr:longint;Value:longint); far; external DLLName;
	function  f4char(FieldPtr:longint):integer; far; external DLLName;
	function  f4decimals(FieldPtr:longint):integer; far; external DLLName;
	function  f4double(FieldPtr:longint):double; far; external DLLName;
	function  f4int(FieldPtr:longint):integer; far; external DLLName;
	function  f4len(FieldPtr:longint):integer; far; external DLLName;
	function  f4long(FieldPtr:longint):longint; far; external DLLName;
	function  f4ncpy(FieldPtr:longint; Erg:PChar; Lng:integer):integer; far; external DLLName;
	function  f4str(FieldPtr:longint):longint; far; external DLLName;
	function  f4true(FieldPtr:longint):integer; far; external DLLName;
	function  f4type(FieldPtr:longint):integer; far; external DLLName;

	function  tag4alias(TagPtr:longint):longint; far; external DLLName;
	function  tag4Descending(TagPtr:longint):integer; far; external DLLName;
	function  tag4expr(TagPtr:longint):longint; far; external DLLName;
	function  tag4filter(TagPtr:longint):longint; far; external DLLName;
	function  tag4KeyLen(TagPtr:longint):integer; far; external DLLName;
	function  tag4unique(TagPtr:longint):integer; far; external DLLName;
	function  tag4uniqueError(TagPtr:longint;Value:integer):longint; far; external DLLName;

	function  u4ncpy(strPtr1:pChar;memPtr2:longint;len:integer):integer; far; external DLLName;

	{ *** Privat Functions *** }

	procedure TCBServer.Loaded;
	var
		Datum: array [0..19] of char;
	begin
		inherited Loaded;
		CODE4 := d4init_v;
		code4autoOpen(CODE4, ord(FAutoOpen));
		code4createError(CODE4, ord(FCreateError));
		case FDateFormat of
			dtGerman: StrCopy(Datum, 'DD.MM.YY');
			dtAmerican: StrCopy(Datum, 'MM/DD/YY');
			dtBritish: StrCopy(Datum, 'MM.DD.YY');
		end;
		code4dateFormat(CODE4, Datum);
		code4fieldNameError(CODE4, ord(FFieldNameError));
		code4goError(CODE4, ord(FgoError));
		code4openError(CODE4, ord(FOpenError));
		code4safety(CODE4, ord(FSafety));
		code4skipError(CODE4, ord(FSafety));
	end;


	procedure TCBFile.SetFileName(Value:string);
	var
		FileName: string;
		PosPkt: byte;
	begin
		FFileName := Value;
		FileName := ExtractFileName(FFileName);
		PosPkt := Pos('.', FileName)-1;
		if PosPkt > 0 then
			FAlias := Copy(FileName, 1, PosPkt)
		else
			FAlias := FileName;
	end;


	{  set Properties in TCBFile }
	procedure TCBFile.SetAllProperties;
	begin
		RecCount := d4reccount(FilePtr);
		RecNum := d4recno(FilePtr);
		EOF := (d4eof(FilePtr)<>0);
		BOF := (d4bof(FilePtr)<>0);
		Deleted := (d4deleted(FilePtr)<>0);
		Locked := (d4lock_test(FilePtr, RecNum)<>0);
	 end;


	{ *** TCBServer Methoden *** }

	function TCBServer.GetCODE4:longint;
	begin
		Result := CODE4;
	end;


	{ *** TCBFile Methoden *** }

	function TCBFile.Append:integer;
	begin
		Result := d4append_blank(FilePtr);
		SetAllProperties;
	end;


	function TCBFile.Close:integer;
	begin
		Result := d4close(FilePtr);
	end;


	procedure TCBFile.Delete;
	begin
		d4delete(FilePtr);
		SetAllProperties;
	end;


	function TCBFile.FileLock:integer;
	begin
		Result := d4lock_file(FilePtr);
	end;


	function TCBFile.GetAlias:string;
	begin
		Result := StrPas(d4Alias(FilePtr));
	end;


	function TCBFile.GetDATA4:longint;
	begin
		Result := FilePtr;
	end;


	function TCBFile.GetError:integer;
	begin
		Result := code4errorCode(FServer.CODE4,0);
	end;


	function TCBFile.GetPosition:integer;
	var
		erg: double;
	begin
		erg := d4position(FilePtr);
		Result := Round(erg*100);
	end;


	function TCBFile.Go(RecNum:longint):integer;
	begin
		Result := d4go(FilePtr, RecNum);
		SetAllProperties;
	end;


	function TCBFile.GoBottom:integer;
	begin
		Result := d4bottom(FilePtr);
		SetAllProperties;
	end;


	function TCBFile.GoEof:integer;
	begin
		Result := d4goEof(FilePtr);
		SetAllProperties;
	end;


	function TCBFile.GoTop:integer;
	begin
		Result := d4top(FilePtr);
		SetAllProperties;
	end;

	function TCBFile.Open:integer;
	var
		erg, tagPtr: longint;

	begin
		TagPtr := 0;
		FilePtr := 0;
		if not (FServer is TCBServer) then
		begin
			Result := cbNoServer;
			exit;
		end;

		code4exclusive(FServer.CODE4, ord(FExclusive));
		code4readLock(FServer.CODE4, ord(FReadLock));
		code4readOnly(FServer.CODE4, ord(FReadOnly));

		StrPCopy(NullStr, FFileName);
		FilePtr := d4open(FServer.CODE4, NullStr);
		{ Error ?? }
		if FilePtr <= 0 then
		begin
			Result := code4errorCode(FServer.CODE4,0);
			Exit;
		end;
		StrPCopy(NullStr, FAlias);
		d4alias_Set(FilePtr, NullStr);

		{ Set Tag }
		if Length(IndexTag)>0 then
		begin
			StrPCopy(NullStr, IndexTag);
			tagPtr := d4tag(FilePtr, NullStr);
		end;
		if tagPtr > 0 then
			d4tag_Select(FilePtr, tagPtr);

		d4top(FilePtr);
		SetAllProperties;
		Fields := d4num_fields(FilePtr);
		Result := 1;
	end;


	function TCBFile.Pack:integer;
	var
		erg:integer;
	begin
		erg := d4pack(FilePtr);
		{ Error ?? }
		if erg = 0 then
			erg := d4memo_compress(FilePtr);
			{ Error ?? }
		Result := erg;
	end;


	procedure TCBFile.Recall;
	begin
		d4recall(FilePtr);
		SetAllProperties;
	end;


	function TCBFile.RecLock:integer;
	begin
		Result := d4lock(FilePtr, RecNum);
	end;


	function TCBFile.ReIndex:integer;
	begin
		Result := d4reindex(FilePtr);
		{ Error ?? }
	end;


	function TCBFile.SetPosition(Position:double):integer;
	begin
		Result := d4position_set(FilePtr, Position);
	end;


	function TCBFile.Seek(Value:string):boolean;
	var
		erg:integer;
	begin
		StrPCopy(NullStr, Value);
		erg := d4Seek(FilePtr, NullStr);
		{ Error ?? }
		if erg > 0 then
			Result := True;
		SetAllProperties;
	end;


	function TCBFile.Skip(NumRecs:longint):integer;
	begin
		Result := d4skip(FilePtr,NumRecs);
		SetAllProperties;
	end;


	function TCBFile.Zap(StartRec:longint;EndRec:longint):integer;
	var
		erg:integer;
	begin
		erg := d4zap(FilePtr, StartRec, EndRec);
		{ Error ?? }
		if erg = 0 then
			erg := d4memo_compress(FilePtr);
			{ Error ?? }
		Result := erg;
	end;


	constructor TCBServer.Create(AOwner: TComponent);
	begin
		inherited Create(AOwner);
		FAutoOpen := True;
		FCreateError := True;
		FDateFormat := dtGerman;
		FFieldNameError := True;
		FGoError := True;
		FOpenError := True;
		FSafety := True;
		FSkipError := True;
	end;



	constructor TCBFile.Create(AOwner: TComponent);
	begin
		inherited Create(AOwner);
		FExclusive := False;
		FReadLock := False;
		FReadOnly := False;
	end;


	{ *** Index(Tag) Methoden *** }

	function TCBFile.fTagPtr(TagName:string):longint;
	begin
		if Length(TagName) > 0 then
		begin
			StrPCopy(NullStr, TagName);
			Result := d4tag(FilePtr, NullStr);
		end;
	end;


	function TCBFile.GetTag:string;
	var
		rc: integer;
		TagPtr: longint;
		TagName: array[0..11] of Char;
	begin
		TagPtr := d4tag_Selected(FilePtr);
		rc := u4ncpy(TagName, TagPtr, 11);
		Result := StrPas(TagName);
	end;


	function TCBFile.GetTagPtr:longint;
	begin
		Result := d4tag_Selected(FilePtr);
	end;


	function TCBFile.GetNextTagPtr(onTag:longint):longint;
	begin
		Result := d4tag_Next(FilePtr, onTag);
	end;


	function TCBFile.GetPrevTagPtr(onTag:longint):longint;
	begin
		Result := d4tag_Prev(FilePtr, onTag);
	end;


	function TCBFile.SetTag(TagName:string):boolean;
	var
		TagPtr:longint;
	begin
		TagPtr := fTagPtr(TagName);
		if TagPtr > 0 then
			d4tag_Select(FilePtr, TagPtr);
		d4top(FilePtr);
		SetAllProperties;
	end;


	function TCBFile.TagDescending(TagName:string):boolean;
	var
		TagPtr:longint;
	begin
		TagPtr := fTagPtr(TagName);
		if TagPtr > 0 then
		begin
			StrPCopy(NullStr, TagName);
			TagPtr := d4tag(FilePtr, NullStr);
			Result := (tag4Descending(TagPtr)=1);
		end;
	end;


	function TCBFile.TagExpr(TagName:string):string;
	var
		rc:integer;
		exprPtr,TagPtr:longint;
	begin
		Result := '';
		TagPtr := fTagPtr(TagName);
		if TagPtr > 0 then
		begin
			exprPtr := tag4expr(TagPtr);
			if exprPtr > 0 then
			begin
				rc := u4ncpy(NullStr, exprPtr, 255);
				Result := StrPas(NullStr);
			end;
		end;
	end;


	function TCBFile.TagFilter(TagName:string):string;
		var
		rc:integer;
		exprPtr,TagPtr:longint;
	begin
		Result := '';
		TagPtr := fTagPtr(TagName);
		if TagPtr > 0 then
		begin
			exprPtr := tag4filter(TagPtr);
			if exprPtr > 0 then
			begin
				rc := u4ncpy(NullStr, exprPtr, 255);
				Result := StrPas(NullStr);
			end;
		end;
	end;


	function TCBFile.TagKeyLen(TagName:string):integer;
	var
		TagPtr:longint;
	begin
		Result := 0;
		TagPtr := fTagPtr(TagName);
		if TagPtr > 0 then
		begin
			Result := tag4KeyLen(TagPtr);
		end;
	end;


	function TCBFile.TagUnique(TagName:string):boolean;
	var
		TagPtr:longint;
	begin
		Result := False;
		TagPtr := fTagPtr(TagName);
		if TagPtr > 0 then
		begin
			Result := (tag4Unique(TagPtr)=1);
		end;
	end;


	function TCBFile.GetTagUniqueError(TagName:string):boolean;
	var
		TagPtr:longint;
	begin
		Result := False;
		TagPtr := fTagPtr(TagName);
		if TagPtr > 0 then
		begin
			Result := (tag4UniqueError(TagPtr,-1)=1);
		end;
	end;


	procedure TCBFile.SetTagUniqueError(TagName:string;Value:boolean);
	var
		TagPtr:longint;
	begin
		TagPtr := fTagPtr(TagName);
		if TagPtr > 0 then
		begin
			if Value = True then
				tag4UniqueError(TagPtr,1)
			else
				tag4UniqueError(TagPtr,0);
		end;
	end;


	{ *** Field Methoden *** }

	function TCBFile.FieldDec(Field:string):Byte;
	begin
		Result := Byte(f4decimals(GetFieldPtr(Field)));
	end;


	function TCBFile.FieldLen(Field:string):Byte;
	begin
		Result := Byte(f4len(GetFieldPtr(Field)));
	end;


	function TCBFile.FieldType(Field:string):Byte;
	begin
		Result := Byte(f4type(GetFieldPtr(Field)));
	end;


	{ Get Field-Pointer for Field }
	function TCBFile.GetFieldPtr(Field:string):longint;
	var
		FieldOrd, Code:integer;
	begin
		Val(Field, FieldOrd, Code);
		if FieldOrd > 0 then
		begin
			Result := d4field_J(FilePtr, FieldOrd);
		end
		else
		begin
			StrPCopy(NullStr, Field);
			Result := d4field(FilePtr, NullStr);
		end;
	end;


	function TCBFile.GetDouble(Field:string):double;
	begin
		Result := f4double(GetFieldPtr(Field));
	end;


	function TCBFile.GetInt(Field:string):integer;
	begin
		Result := f4int(GetFieldPtr(Field));
	end;


	function TCBFile.GetLong(Field:string):longint;
	begin
		Result := f4long(GetFieldPtr(Field));
	end;


	function TCBFile.GetBoolean(Field:string):boolean;
	begin
		Result := (f4true(GetFieldPtr(Field)) = 1 );
	end;


	function TCBFile.GetStr(Field:string):string;
	begin
		f4ncpy( GetFieldPtr(Field), NullStr, SizeOf(NullStr));
		Result := StrPas(NullStr);
	end;


	procedure TCBFile.SetBoolean(Field:string;Value:boolean);
	begin
		if Value = True then
			f4assign_Char(GetFieldPtr(Field), Ord('T'))
		else
			f4assign_Char(GetFieldPtr(Field), Ord('F'));
	end;


	procedure TCBFile.SetDouble(Field:string;Value:double);
	begin
		f4assign_Double(GetFieldPtr(Field),Value);
	end;


	procedure TCBFile.SetInt(Field:string;Value:integer);
	begin
		f4assign_Int(GetFieldPtr(Field),Value);
	end;


	procedure TCBFile.SetLong(Field:string;Value:longint);
	begin
		f4assign_Long(GetFieldPtr(Field),Value);
	end;


	procedure TCBFile.SetStr(Field:string;Value:string);
	begin
		StrPCopy(NullStr, Value);
		f4assign(GetFieldPtr(Field),NullStr);
	end;



initialization

end.

